/* 8BitAmEthernet
 * version 0.1 date 2005-02-21
 * Copyright 2005 Stefan Schuermans <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 * a project of CCC-AC - http://www.cccac.de/
 */

#include <avr/interrupt.h>

#include "code.h"
#include "config.h"
#include "eeprom.h"
#include "macros.h"
#include "rtl8019.h"

//convert 3 BCD digits to a byte
static inline unsigned char Code3Bcd2Byte( unsigned char * pCode, unsigned char Pos )
{
  //byte starts at even position
  if( (Pos & 0x01) == 0x00 )
    return (pCode[Pos>>1] >> 4) * 100 + (pCode[Pos>>1] & 0x0F) * 10 + (pCode[(Pos>>1)+1] >> 4);
  //byte starts at odd position
  else
    return (pCode[Pos>>1] & 0x0F) * 100 + (pCode[(Pos>>1)+1] >> 4) * 10 + (pCode[(Pos>>1)+1] & 0x0F);
}

//set MAC address
//code is stored in BCD, MSB first - magic is already removed
//returns 0x00 on success
static inline unsigned char CodeSetMac( unsigned char * pCode, unsigned char CodeLen )
{
  unsigned char b4, b5;

  //invalid length
  if( CodeLen != 8 )
    return 0xFF;

  //get last 2 bytes from code
  b4 = Code3Bcd2Byte( pCode, 2 );
  b5 = Code3Bcd2Byte( pCode, 5 );

  //set MAC address to 02:CC:C0:AC:__:__ (atomically)
  cli( );
  ConfigMac[0] = 0x02;
  ConfigMac[1] = 0xCC;
  ConfigMac[2] = 0xC0;
  ConfigMac[3] = 0xAC;
  ConfigMac[4] = b4;
  ConfigMac[5] = b5;
  sei( );
  //schedule write to EEPROM
  EepromNeedWrite |= EepromNeedWriteMac;

  //re-initialize RTL8019
  cli( );
  RtlReinit( );
  sei( );

  return 0x00;
}

//set IP address
//code is stored in BCD, MSB first - magic is already removed
//returns 0x00 on success
static inline unsigned char CodeSetIp( unsigned char * pCode, unsigned char CodeLen )
{
  unsigned char b0, b1, b2, b3;

  //invalid length
  if( CodeLen != 14 )
    return 0xFF;

  //get 4 bytes from code
  b0 = Code3Bcd2Byte( pCode, 2 );
  b1 = Code3Bcd2Byte( pCode, 5 );
  b2 = Code3Bcd2Byte( pCode, 8 );
  b3 = Code3Bcd2Byte( pCode, 11 );

  //set IP address (atomically)
  cli( );
  ConfigIp[0] = b0;
  ConfigIp[1] = b1;
  ConfigIp[2] = b2;
  ConfigIp[3] = b3;
  sei( );
  //schedule write to EEPROM
  EepromNeedWrite |= EepromNeedWriteIp;

  return 0x00;
}

//set subnet mask
//code is stored in BCD, MSB first - magic is already removed
//returns 0x00 on success
static inline unsigned char CodeSetMask( unsigned char * pCode, unsigned char CodeLen )
{
  unsigned char b[4];
  unsigned long l;

  //invalid length
  if( CodeLen != 14 )
    return 0xFF;

  //get 4 bytes from code
  b[0] = Code3Bcd2Byte( pCode, 2 );
  b[1] = Code3Bcd2Byte( pCode, 5 );
  b[2] = Code3Bcd2Byte( pCode, 8 );
  b[3] = Code3Bcd2Byte( pCode, 11 );

  //check subnet mask (subnet mask must be 1..10..0 in binary)
  l = (unsigned long)b[0] << 24 | (unsigned long)b[1] << 16 | (unsigned long)b[2] << 8 | (unsigned long)b[3];
  while( l & 0x80000000 ) //shift out 1 bits
    l <<= 1;
  if( l != 0x00000000 ) //if not every bit is 0 ...
    return 0xFF; //... there must have been a 1 bit after a 0 bit (invalid netmask)

  //set subnet mask (atomically)
  cli( );
  ConfigMask[0] = b[0];
  ConfigMask[1] = b[1];
  ConfigMask[2] = b[2];
  ConfigMask[3] = b[3];
  sei( );
  //schedule write to EEPROM
  EepromNeedWrite |= EepromNeedWriteMask;

  return 0x00;
}

//set gateway IP address
//code is stored in BCD, MSB first - magic is already removed
//returns 0x00 on success
static inline unsigned char CodeSetGw( unsigned char * pCode, unsigned char CodeLen )
{
  unsigned char b0, b1, b2, b3;

  //invalid length
  if( CodeLen != 14 )
    return 0xFF;

  //get 4 bytes from code
  b0 = Code3Bcd2Byte( pCode, 2 );
  b1 = Code3Bcd2Byte( pCode, 5 );
  b2 = Code3Bcd2Byte( pCode, 8 );
  b3 = Code3Bcd2Byte( pCode, 11 );

  //set gateway IP address (atomically)
  cli( );
  ConfigGw[0] = b0;
  ConfigGw[1] = b1;
  ConfigGw[2] = b2;
  ConfigGw[3] = b3;
  sei( );
  //schedule write to EEPROM
  EepromNeedWrite |= EepromNeedWriteGw;

  return 0x00;
}

//set symmetric key
//code is stored in BCD, MSB first - magic is already removed
//returns 0x00 on success
static inline unsigned char CodeSetKey( unsigned char * pCode, unsigned char CodeLen )
{
  unsigned char key[16], i, j;

  //invalid length
  if( CodeLen != 2 + count( key ) * 3 )
    return 0xFF;

  //get bytes from code
  for( i = 0, j = 2; i < count( key ); i++, j += 3 )
    key[i] = Code3Bcd2Byte( pCode, j );

  //set symmetric key (atomically)
  cli( );
  for( i = 0; i < min( count( key ), count( ConfigKey ) ); i++ )
    ConfigKey[i] = key[i];
  sei( );
  //schedule write to EEPROM
  EepromNeedWrite |= EepromNeedWriteKey;

  return 0x00;
}

//deactivate webserver
//code is stored in BCD, MSB first - magic is already removed
//returns 0x00 on success
static inline unsigned char CodeDeactWeb( unsigned char * pCode, unsigned char CodeLen )
{
  //invalid length
  if( CodeLen != 2 )
    return 0xFF;

  //deactivate webserver (atomically because only a single assignment)
  ConfigWeb = 0x00;
  //schedule write to EEPROM
  EepromNeedWrite |= EepromNeedWriteWeb;

  return 0x00;
}

//activate webserver
//code is stored in BCD, MSB first - magic is already removed
//returns 0x00 on success
static inline unsigned char CodeActWeb( unsigned char * pCode, unsigned char CodeLen )
{
  //invalid length
  if( CodeLen != 2 )
    return 0xFF;

  //activate webserver (atomically because only a single assignment)
  ConfigWeb = 0x01;
  //schedule write to EEPROM
  EepromNeedWrite |= EepromNeedWriteWeb;

  return 0x00;
}

//process a code
//code is stored in BCD, MSB first
//returns 0x00 on success
unsigned char CodeProc( unsigned char * pCode, unsigned char CodeLen )
{
  //check green LED
  if( CodeLen == 0 )
    return 0x00;

  //not long enough for magic
  if( CodeLen < 2 )
    return 0xFF;

  //branch according to magic
  switch( pCode[0] )
  {
    case 0x00: return CodeSetMac( pCode, CodeLen ); //set MAC address
    case 0x10: return CodeSetIp( pCode, CodeLen ); //set IP address
    case 0x11: return CodeSetMask( pCode, CodeLen ); //set subnet mask address
    case 0x12: return CodeSetGw( pCode, CodeLen ); //set gateway IP address
    case 0x20: return CodeSetKey( pCode, CodeLen ); //set symmetric key
    case 0x30: return CodeDeactWeb( pCode, CodeLen ); //deactivate webserver
    case 0x31: return CodeActWeb( pCode, CodeLen ); //activate webserver
  }

  //unknown code
  return 0xFF;
}

